########################################################################
#                 Copyright (c) 1997 by IV DocEye AB
#              Copyright (c) 1999 by Stephen J. Turner
#                Copyright (c) 2005 by Carsten Haese
#  
# By obtaining, using, and/or copying this software and/or its
# associated documentation, you agree that you have read, understood,
# and will comply with the following terms and conditions:
#  
# Permission to use, copy, modify, and distribute this software and its
# associated documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appears in all
# copies, and that both that copyright notice and this permission notice
# appear in supporting documentation, and that the name of the author
# not be used in advertising or publicity pertaining to distribution of
# the software without specific, written prior permission.
#  
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.  IN
# NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
# USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
########################################################################

# $Id$
# informixdb.py
#
# This is a trivial python wrapper around the C core _informixdb.so
#

"""\
 DB-API 2.0 compliant interface for IBM Informix databases.

Here's a small example to get you started:

>>> import informixdb
>>> conn = informixdb.connect('mydatabase')
>>> cursor = conn.cursor()
>>> cursor.execute('SELECT * FROM names')
>>> cursor.fetchall()
[('donald', 'duck', 34), ('mickey', 'mouse', 23)]

For more information on DB-API 2.0, see
http://www.python.org/peps/pep-0249.html
"""

version = "2.5"

class Row(object):
  """Helper class for cursors whose row format is ROW_AS_OBJECT."""
  def __init__(self, d): self.__dict__.update(d)
  def __repr__(self): return repr(self.__dict__)
  def __str__(self): return str(self.__dict__)
  def __getitem__(self,k): return self.__dict__[k]

class IntervalYearToMonth(object):
  """\
This class is used for input and output binding of
INTERVAL columns whose precision is a subset of
YEAR TO MONTH.

Intervals can be added to dates, datetimes, and other Intervals
that have compatible precision. Intervals can also be multiplied
and divided by scalar factors.
"""
  def __init__(self, years=0, months=0):
    self._months = 12*years+months
    self.years, self.months = divmod(self._months,12)
  def __repr__(self):
    return "%s(%d, %d)"%(self.__class__.__name__,self.years,self.months)
  def __str__(self):
    """Returns the interval in Informix's format for input binding."""
    if self._months < 0:
      return "-%d-%02d" % divmod(-self._months,12)
    else:
      return "%d-%02d" % divmod(self._months,12)
  def __add__(self, other):
      if isinstance(other, IntervalYearToMonth):
        return self.__class__(0, int(self._months+other._months))
      elif isinstance(other, (datetime.datetime, datetime.date)):
        # extract the date from the other operand
        y,m,d = other.timetuple()[0:3]
        otherdate = datetime.date(y,m,d)
        # shift the date by the desired number of months
        y2,m2 = divmod(m-1+self._months,12)
        try:
          date2 = datetime.date(y+y2,m2+1,d)
        except ValueError:
          raise ValueError("month arithmetic yielded an invalid date.")
        # apply the resulting timedelta to the operand
        return other + (date2 - otherdate)
      else: return NotImplemented
  def __neg__(self): return self.__class__(0, -self._months)
  def __sub__(self, other): return self + -other
  def __rsub__(self, other): return -self + other
  def __abs__(self):
      if self._months < 0: return -self
      else: return self
  def __mul__(self, other):
      #if isinstance(other, (int,long,float)):
      if isinstance(other, (int,float)):
        return self.__class__(0, int(self._months*other))
      else: return NotImplemented
  def __div__(self, other):
      #if isinstance(other, (int,long,float)):
      if isinstance(other, (int,float)):
        return self.__class__(0, int(self._months/other))
      else: return NotImplemented
  __radd__ = __add__
  __rmul__ = __mul__
  __floordiv__ = __div__
  def __cmp__(self, other):
      """Implements comparisons between intervals."""
      if isinstance(other, IntervalYearToMonth):
        return self._months - other._months
      else: return NotImplemented

# Define IntervalDayToFraction for symmetry. All the heavy lifting is done
# by datetime.timedelta, from which this class is derived.
import datetime
class IntervalDayToFraction(datetime.timedelta):
  """\
This class is used for input and output binding of
INTERVAL columns whose precision is a subset of
DAY TO FRACTION.

Intervals can be added to dates, datetimes, and other Intervals
that have compatible precision. Intervals can also be multiplied
and divided by scalar factors.
"""
  def __init__(self,days=0,seconds=0,microseconds=0):
    datetime.timedelta.__init__(self, days, seconds, microseconds)
  def __str__(self):
    """Returns the interval in Informix's format for input binding."""
    if self.days<0:
      neg = IntervalDayToFraction(-self.days, -self.seconds, -self.microseconds)
      return '-'+str(neg)
    else:
      minutes, seconds = divmod(self.seconds, 60)
      hours, minutes = divmod(minutes, 60)
      return "%d %02d:%02d:%02d.%05d" % (self.days, hours,minutes,seconds,
                                         self.microseconds/10)

# The module initialization of _informixdb references the Interval classes
# above, so don't put this import before the interval class definitions.
from _informixdb import *

# promote the class definitions of _informixdb.Cursor and
# _informixdb.Connection into this namespace so that help(informixdb)
# sees their doc strings. Also make it impossible for the user to try
# to instantiate these classes directly.
from _informixdb import Cursor as _Cursor, Connection as _Connection
class Cursor(_Cursor):
  __doc__ = _Cursor.__doc__
  def __new__(self, *args, **kwargs):
    raise InterfaceError("Use Connection.cursor() to instantiate a cursor.")
del _Cursor
class Connection(_Connection):
  __doc__ = _Connection.__doc__
  def __new__(self, *args, **kwargs):
    raise InterfaceError("Use connect() to instantiate a connection.")
del _Connection
try:
  # Same for Sblobs if we have support for them in _informixdb
  from _informixdb import Sblob as _Sblob
  class Sblob(_Sblob):
    __doc__ = _Sblob.__doc__
    def __new__(self, *args, **kwargs):
      raise InterfaceError("Use Connection.Sblob() to instantiate an Sblob.")
  del _Sblob
except: pass



import sys
import os
import shlex
import re
from distutils.spawn import find_executable
from distutils.sysconfig import get_python_inc
from distutils.util import get_platform
from distutils.command.build_ext import build_ext as _build_ext
from distutils.command.bdist_wininst import bdist_wininst as _bdist_wininst
from distutils.dep_util import newer_group
from distutils.errors import *

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

class build_ext(_build_ext):
    """ build_ext which can handle ESQL/C (*.ec) files """

    user_options = _build_ext.user_options + [
        ('esql-threadlib=', None,
        '[ESQL/C] Thread library to use with ESQL/C'),
        ('esql-informixdir=', None,
        '[ESQL/C] Informixdir to use if $INFORMIXDIR is not set'),
        ('esql-static', None,
        '[ESQL/C] statically link against ESQL/C libraries')
        ]

    boolean_options = [ 'esql-static' ]

    def initialize_options(self):
        _build_ext.initialize_options(self)

        self.esql_informixdir = None
        self.esql_threadlib = None
        self.esql_static = 0 # link staticly by default
        self.esql_parts = []

    def finalize_options(self):
        _build_ext.finalize_options(self)

        if not self.esql_informixdir:
            self.esql_informixdir = os.getenv("INFORMIXDIR")
        if not self.esql_informixdir:
          if get_platform()=="win32":
            self.esql_informixdir = "C:\\Program Files\\Informix\\Client-SDK"
          else:
            self.esql_informixdir = "/usr/informix"
        os.environ['INFORMIXDIR'] = self.esql_informixdir

        self.esql_parts.append(os.path.join(self.esql_informixdir,'bin','esql'))
        if get_platform()=="win32":
          if self.esql_parts[0].find(' ') != -1:
            self.esql_parts[0] = '"' + self.esql_parts[0] + '"'

        if self.esql_threadlib:
            os.environ['THREADLIB'] = self.esql_threadlib
            self.esql_parts.append('-thread')
            if self.define is None:
                self.define = []
            self.define += [ ('IFX_THREAD',None), ('_REENTRANT',None) ]

        if self.esql_static:
            self.esql_parts.append('-static')
        self.esql_parts.append('-g -O0')

        # determine esql version
        driver_name = "INFORMIX-ESQL"
        driver_version = "Unknown"
        esqlver = re.compile(r"(IBM)?.*ESQL Version ((\d+)\.(\d+)[^ ]*)")
        cout = os.popen(' '.join(self.esql_parts[0:1] + [ '-V' ]),'r')
        esqlversion = None
        for line in cout:
          matchobj = esqlver.match(line)
          if matchobj:
            matchgroups = matchobj.groups()
            driver_version = matchgroups[1].strip()
            esqlversion = int(matchgroups[2] + matchgroups[3])
            if matchgroups[0]=="IBM":
              # Assume ESQL 9.xx for any IBM branded CSDK.
              driver_name = "IBM Informix-ESQL"
              esqlversion = 960
        esqlversion = 1210
        if esqlversion==None:
          esqlversion = 850
        if esqlversion >= 900:
          self.esql_parts.append("-EDHAVE_ESQL9")
        if esqlversion >= 953:
          self.esql_parts.append("-EDHAVE_DESCRIBE_INPUT")
        f = open(os.path.join("ext","esqlver.h"), "w")
        f.write("""\
#define DRIVER_NAME "%(driver_name)s"
#define DRIVER_VERSION "%(driver_version)s"
""" % locals())
        f.close()

        # find esql libs/objects
        cout = os.popen(' '.join(self.esql_parts + [ '-libs' ]),'r')
        esql_config = []
        lexer = shlex.shlex(cout)
        lexer.wordchars += '-.\\/'
        while True:
          token = lexer.get_token()
          if token=='' or token==None: break
          if token.startswith('"') and token.endswith('"'):
            token = token[1:-1]
          esql_config.append(token)
        ret = cout.close()
        if ret != None:
          raise DistutilsSetupError("""\
Can't run esql. Please make sure that:
* You have the Informix CSDK installed,
* INFORMIXDIR is set to where Informix CSDK is installed, and
* esql is in your PATH.

See the README for build requirements.
""")

        if get_platform()=="win32":
          for arg in esql_config:
              if arg.endswith('.lib'):
                  if self.libraries is None:
                      self.libraries = []
                  self.libraries.append(arg[:-4])
        else:
          for arg in esql_config:
              if arg.startswith('-l'):
                  if self.libraries is None:
                      self.libraries = []
                  self.libraries.append(arg[2:])
              else:
                  if self.link_objects is None:
                      self.link_objects = []
                  self.link_objects.append(arg)

        if self.include_dirs is None:
            self.include_dirs = []
        self.include_dirs = [os.path.join(self.esql_informixdir,'incl','esql')]\
                            + self.include_dirs

        if self.library_dirs is None:
            self.library_dirs = []
        self.library_dirs += [os.path.join(self.esql_informixdir,'lib','esql'),
                              os.path.join(self.esql_informixdir,'lib')]

    def build_extension(self, ext):
        # only preprocess with esql if necessary
        fullname = self.get_ext_fullname(ext.name)
        ext_filename = os.path.join(self.build_lib,
                                        self.get_ext_filename(fullname))
        if not (self.force or newer_group(ext.sources, ext_filename, 'newer')):
            self.announce("skipping '%s' extension (up-to-date)" % ext.name)
            return

        # preprocess *.ec files with 'esql'
        for file in ext.sources:
            if file.endswith('.ec'):
                dir = os.path.dirname(file)
                f = os.path.basename(file)
                cmd = ' '.join(self.esql_parts + [ '-e', f ])
                print(cmd)

                curdir = os.getcwd()
                os.chdir(dir)
                os.system(cmd)
                os.chdir(curdir)

                ext.sources[ext.sources.index(file)] = file[:-3]+'.c'

        _build_ext.build_extension(self, ext)

class bdist_wininst(_bdist_wininst):
    """ override bdist_wininst to include the license in the description. """
    def get_inidata(self):
        metadata = self.distribution.metadata
        save_long_desc = metadata.long_description
        metadata.long_description += "\n\n" + file("COPYRIGHT").read() + "\n"
        result = _bdist_wininst.get_inidata(self)
        metadata.long_description = save_long_desc
        return result

def have_c_datetime():
    """ Check whether the datetime C API is available. """
    v = sys.version_info
    if sys.version_info[0] > 2:
        return 1
    elif sys.version_info[0] == 2 and sys.version_info[1] >= 4:
        return 1
    else:
        return 0

def have_py_bool():
    if sys.version_info[0] > 2:
        return 1
    elif sys.version_info[0] == 2 and sys.version_info[1] >= 3:
        return 1
    else:
        return 0

extra_macros = [('PYTHON_INCLUDE', get_python_inc(plat_specific=1)),
                ('HAVE_C_DATETIME', have_c_datetime()),
                ('HAVE_PY_BOOL', have_py_bool()) ]

modules = [ 'informixdb' ]

# If we don't have a datetime module available, install ours
saved_path = sys.path
try:
    sys.path = sys.path[1:]
    import datetime
except:
    modules.append('datetime')
sys.path = saved_path

# On AIX we need to define _H_LOCALEDEF, so that the system's
# loc_t doesn't conflict with Informix' loc_t for BLOBs
if get_platform().startswith('aix-'):
  extra_macros.append(('_H_LOCALEDEF', None))

module1 = Extension('_informixdb',
                    sources = [os.path.join('ext','_informixdb.ec')],
                    include_dirs = ['ext'],
                    define_macros = extra_macros )

# patch distutils if it can't cope with the "classifiers" or 
# "download_url" keywords
if sys.version < '2.2.3':
    from distutils.dist import DistributionMetadata
    DistributionMetadata.classifiers = None
    DistributionMetadata.download_url = None

setup (name = 'InformixDB-Knight',
       version = '3.0.0.3',
       description = 'InformixDB v3.0',
       long_description = \
         "InformixDB is a DB-API 2.0 compliant interface for IBM Informix Work For Python3\n"
         "databases.",
       url = "http://sourceforge.net/projects/informixdb",
       license = "BSD License",
       platforms = ["POSIX", "Microsoft Windows 95/98/NT/2000/XP"],
       classifiers=[
         "Development Status :: 5 - Production/Stable",
         "Environment :: Console",
         "Intended Audience :: Developers",
         "License :: OSI Approved :: BSD License",
         "Operating System :: Microsoft :: Windows :: Windows 95/98/2000",
         "Operating System :: Microsoft :: Windows :: Windows NT/2000",
         "Operating System :: POSIX",
         "Programming Language :: C",
         "Programming Language :: Python",
         "Topic :: Database :: Front-Ends"
       ],
       py_modules = modules,
       ext_modules = [module1],
       cmdclass = {
         'build_ext' : build_ext,
         'bdist_wininst': bdist_wininst
       } )
